//-----------------------------------------------------------------------------
// File:     TwoPendulumsConnectedBySpringDamperOrConstraint.cpp
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simple example of using a point-to-point "distance" constraint or a spring/damper.
// Authors:  Michael Sherman and Paul Mitiguy -  June 1, 2007
//-----------------------------------------------------------------------------
//#include "StandardCppHeadersAndNamespace.h"
#include "SimTKsimbody.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  DoRequiredTasks( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != EOF; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteStringToScreenAndFile( const char s[], FILE *fptr )      { return WriteStringToScreen( s ) && WriteStringToFile( s, fptr ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
bool  WriteDoubleToScreenAndFile( double x, int precision, FILE *fptr )  { return WriteDoubleToFile( x, precision, stdout )  &&  WriteDoubleToFile( x, precision, fptr ); }
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );
Real  ConvertFromRadiansToDegrees( Real angleInRadians )  { return angleInRadians * 57.295779513082320876798154814105170332405472466564321549160243861;   }
Real  ConvertFromDegreesToRadians( Real angleInDegrees )  { return angleInDegrees * 0.017453292519943295769236907684886127134428718885417254560971914402; }
//void  SleepUntilThisClockTime( clock_t waitUntilThisClockTime )  { while( waitUntilThisClockTime > clock() ); }
//void  SleepMilliseconds( long numberOfMillisecondsToSleep )      { SleepUntilThisClockTime( clock() + numberOfMillisecondsToSleep ); }


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
  // Default value is program failed
  bool programSucceeded = false;

  // It is a good programming practice to do little in the main function of a program.
  // The try-catch code in this main routine catches exceptions thrown by functions in the
  // try block, e.g., catching an exception that occurs when a NULL pointer is de-referenced.
  try
  {
     // Do the required programming tasks
     programSucceeded = DoRequiredTasks();
  }
  // This catch statement handles certain types of exceptions
  catch( const exception &e )
  {
     WriteStringToScreen( "\n\n Error: Programming error encountered.\n The exception thrown is: " );
     WriteStringToScreen( e.what() );
  }
  // The exception-declaration statement (...) handles any type of exception,
  // including C exceptions and system/application generated exceptions.
  // This includes exceptions such as memory protection and floating-point violations.
  // An ellipsis catch handler must be the last handler for its try block.
  catch( ... )
  {
     WriteStringToScreen( "\n\n Error: Programming error encountered.\n An unhandled exception was thrown." );
  }

  // Give the user a chance to see a final message before exiting the program.
  WriteStringToScreen( programSucceeded ? "\n\n Program succeeded.  Press  Enter  to finish: " : "\n\n Program failed.  Press  Enter  to finish: " );
  getchar();  // Keeps the screen displayed until the user presses the Enter (or Return) key.

  // The value returned by the main function is the exit status of the program.
  // A normal program exit returns 0 (other return values usually signal an error).
  return programSucceeded == true ? 0 : 1;
}


//-----------------------------------------------------------------------------
bool  DoRequiredTasks( void )
{
  // Declare a multibody system
  // Each multibody system contains at most one SimbodyMatterSubsytem (which inherits from MatterSubsystem - the ma in F=ma)
  // Each multibody system can contain 0, 1, 2, ... force subsystems (the F in F=ma)
  MultibodySystem  mbs;
  SimbodyMatterSubsystem  matterSubsystem;      // Create a single matter sub-system
  mbs.setMatterSubsystem( matterSubsystem );    // Add the matter sub-system to the system.

  // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
  // 1. Create a gravity vector that is straight down (in the ground's frame)
  // 2. Create a uniform gravity sub-system
  // 3. Add the gravity sub-system to the multibody system
  Vec3 gravityVector( 0, -9.8, 0 );
  UniformGravitySubsystem gravity( gravityVector );
  mbs.addForceSubsystem( gravity );

 // Create the mass, center of mass, and inertia properties of both pendulum
  const Real     massOfPendulum = 1.0;
  const Vec3     pendulumMassCenterLocationFromBodyOrigin( 0, 0, 0 );
  const Inertia  pendulumInertiaMatrixAboutBodyOrigin(0);
  MassProperties pendulumMassProperties( massOfPendulum, pendulumMassCenterLocationFromBodyOrigin, pendulumInertiaMatrixAboutBodyOrigin );

  // The Outboard Frame's axes are aligned with the pendulum axes.
  // The Outboard origin       is  offset  from the pendulum's origin in the Pendulum's y-direction.
  const Rotation   outboardFrameOrientationInPendulum;     // Default constructor is identity matrix
  const Vec3       outboardFrameOriginLocationFromPendulumOrigin( 0, 5, 0 );
  const Transform  outboardFrameTransformFromPendulum( outboardFrameOrientationInPendulum, outboardFrameOriginLocationFromPendulumOrigin );

  // The Inboard Frame's axes are aligned    with the Ground's axes.
  // The Inboard origin (left)    is  offset from the Ground's origin by -1.
  // The Inboard origin (right)   is  offset from the Ground's origin by +1.
  const Rotation   inboardFrameOrientationInGround;   // Default constructor is identity matrix
  const Transform  inboardFrameTransformFromGroundLeft(  inboardFrameOrientationInGround, Vec3(-5,0,0) );
  const Transform  inboardFrameTransformFromGroundRight( inboardFrameOrientationInGround, Vec3( 5,0,0) );

  // Connect pendulums to ground
  Mobilizer pendulumToGroundMobilizer = Mobilizer::Pin();
  BodyId leftPendulum  = matterSubsystem.addRigidBody( pendulumMassProperties, outboardFrameTransformFromPendulum, GroundId, inboardFrameTransformFromGroundLeft,  pendulumToGroundMobilizer );
  BodyId rightPendulum = matterSubsystem.addRigidBody( pendulumMassProperties, outboardFrameTransformFromPendulum, GroundId, inboardFrameTransformFromGroundRight, pendulumToGroundMobilizer );

  // Beauty is in the eye of the beholder ...
 /* DecorationSubsystem  viz;
  mbs.setDecorationSubsystem(viz);
  viz.addBodyFixedDecoration( leftPendulum,  Transform(), DecorativeSphere(0.1).setColor(Red)  );
  viz.addBodyFixedDecoration( rightPendulum, Transform(), DecorativeSphere(0.1).setColor(Blue) );
*/

     // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

	// alternate visualization
    const Vec3 brickDimensions(1,1,1);
   DecorativeBrick blueBook = DecorativeBrick(brickDimensions);
   blueBook.setColor(Blue);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   blueBook.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate the ball with the blue sphere at the apple's origin
   const Rotation  bookToBlueBookOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      bookOriginToBlueBookOriginLocation( 0, 0, 0 );
   const Transform bookToBlueBookTransform( bookToBlueBookOrientation, bookOriginToBlueBookOriginLocation );
   animationResults.addDecoration( leftPendulum, bookToBlueBookTransform, blueBook );

   DecorativeBrick redBook = DecorativeBrick(brickDimensions);
   redBook.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   redBook.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate the ball with the blue sphere at the apple's origin
   const Rotation  bookToRedBookOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      bookOriginToRedBookOriginLocation( 0, 0, 0 );
   const Transform bookToRedBookTransform( bookToRedBookOrientation, bookOriginToRedBookOriginLocation );
   animationResults.addDecoration( rightPendulum, bookToRedBookTransform, redBook );


  // Choose to join the two pendulums by a constraint or spring/damper or nothing
  // Add visualization line (orange=spring, black=constraint)
  char connectPendulums = 's';  // Choose c = constraint  or  s = spring  or  something else = noConnection
/*
  if( connectPendulums == 'c' )
  { 
      Real distanceBetweenTwoPoints = 2.0;
      Vec3 constraintOffsetLeftPendulumOrigin(0,0,0);
      Vec3 constraintOffsetRightPendulumOrigin(0,0,0);
      matterSubsystem.addConstantDistanceConstraint( leftPendulum, constraintOffsetLeftPendulumOrigin, rightPendulum, constraintOffsetRightPendulumOrigin, distanceBetweenTwoPoints );

      // For visualization, draw a thin green line
      viz.addRubberBandLine( leftPendulum, constraintOffsetLeftPendulumOrigin, rightPendulum, constraintOffsetRightPendulumOrigin, DecorativeLine().setColor(Black).setLineThickness(7) );
  }
  else if( connectPendulums == 's' ) */
  {
      // Create spring/damper forces and add them to multibody system
      GeneralForceElements forces;
      mbs.addForceSubsystem( forces );

      Real springConstant      = 100;
      Real springNaturalLength = 10;
      Real dampingConstant     = 2.2;
      Vec3 constraintOffsetLeftPendulumOrigin(0,0,0);
      Vec3 constraintOffsetRightPendulumOrigin(0,0,0);
      forces.addTwoPointLinearSpring( leftPendulum, constraintOffsetLeftPendulumOrigin, rightPendulum, constraintOffsetRightPendulumOrigin, springConstant, springNaturalLength );
      forces.addTwoPointLinearDamper( leftPendulum, constraintOffsetLeftPendulumOrigin, rightPendulum, constraintOffsetRightPendulumOrigin, dampingConstant );
	
      // For visualization, draw a thick black line
//      viz.addRubberBandLine( leftPendulum, constraintOffsetLeftPendulumOrigin, rightPendulum, constraintOffsetRightPendulumOrigin, DecorativeLine().setColor(Green).setLineThickness(4) );
  }

  // Create a state for this system.
  // Define appropriate states for this multi-body system.
  // Set the initial time to 0.0
  State s;
  mbs.realize( s); // define appropriate states for this System
  s.setTime( 0.0 );

  // Set the initial values for the configuration variables (x,y,z)
  Real initialAngleInRads = ConvertFromDegreesToRadians( -60 );
  matterSubsystem.setMobilizerQ( s,  leftPendulum, 0,  initialAngleInRads );
  matterSubsystem.setMobilizerQ( s, rightPendulum, 0,  0.0 );

  // Set the initial values for the motion variables
  matterSubsystem.setMobilizerU( s,  leftPendulum, 0,  0.0 );
  matterSubsystem.setMobilizerU( s, rightPendulum, 0,  0.0 );

  // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
  // Set the numerical accuracy for the integrator
  RungeKuttaMerson myStudy( mbs, s );
  myStudy.setAccuracy( 1.0E-7 );

  // Visualize results with VTKReporter
//  VTKReporter animationResults( mbs );

  // Display pendulums before and after assembly analysis has been done
/*  if( connectPendulums == 'c' )
  {
     animationResults.report(s);
     SleepMilliseconds( 1500 );

     myStudy.initialize();   // Does lots of accounting and performs assembly analysis

     animationResults.report(s);
     SleepMilliseconds( 3000 );
  }
  // The next statement does lots of accounting
  else */
	myStudy.initialize();

  // Open a file to record the simulation results (they are also displayed on screen)
  FILE *outputFile = FileOpenWithMessageIfCannotOpen( "TwoPendulumsConnectedBySpringDamperOrConstraint.txt", "w" );
  WriteStringToScreenAndFile( "time    qLeftPendulum   qRightPendulum  mechanicalEnergy\n",  outputFile );

   // Set the numerical integration step and the time for the simulation to run
  const Real integrationStepDt = 0.01;
  const Real finalTime = 5.0;
  const Real finalTimeCompare = finalTime - 0.01*integrationStepDt;

  // Run the simulation and print the results
  while( 1 )
  {
     // Query for results to be printed
     Real time = s.getTime();
     Real leftPendulumAngleRads  = matterSubsystem.getMobilizerQ( s,  leftPendulum, 0 );
     Real rightPendulumAngleRads = matterSubsystem.getMobilizerQ( s, rightPendulum, 0 );
     Real leftPendulumAngleDegs  = ConvertFromRadiansToDegrees(  leftPendulumAngleRads );
     Real rightPendulumAngleDegs = ConvertFromRadiansToDegrees( rightPendulumAngleRads );
     Real kineticEnergy = mbs.getKineticEnergy(s);
     Real potentialEnergy = mbs.getPotentialEnergy(s);
     Real mechanicalEnergy = kineticEnergy + potentialEnergy;

     // Print results to screen and file
     WriteDoubleToScreenAndFile(  time,                  2, outputFile );
     WriteDoubleToScreenAndFile(  leftPendulumAngleDegs, 7, outputFile );
     WriteDoubleToScreenAndFile( rightPendulumAngleDegs, 7, outputFile );
     WriteDoubleToScreenAndFile( mechanicalEnergy,       7, outputFile );
     WriteStringToScreenAndFile( "\n",                      outputFile );

     // Animate the results for this step
     animationResults.report(s);

     // Check if integration has completed
     if( time >= finalTimeCompare ) break;

     // Increment time step
     myStudy.step( time + integrationStepDt );
  }

  // Get the number of steps taken by the numerical integrator
  long numberOfSteps = myStudy.getStepsTaken();

  // Print results to screen and file
  printf( "\n Number of steps taken by numerical integrator %ld: \n", numberOfSteps );

  // Simulation completed properly
  return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
  // Try to open the file
  FILE *Fptr1 = fopen( filename, attribute );

  // If unable to open the file, issue a message
  if( !Fptr1 )
  {
     WriteStringToScreen( "\n\n Unable to open the file: " );
     WriteStringToScreen( filename );
     WriteStringToScreen( "\n\n" );
  }

  return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
  // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
  // Next, calculate the field width so it includes one extra space to the right of the number.
  if( precision < 0 || precision > 17 ) precision = 5;
  int fieldWidth = precision + 8;

  // Create the format specifier and print the number
  char format[20];
  sprintf( format, " %%- %d.%dE", fieldWidth, precision );
  return fprintf( fptr, format, x ) >= 0;
}