//-----------------------------------------------------------------------------
// File:     UserForceTorsionalSpring.h
// Class:    UserForceTorsionalSpring.h
// Parent:   GeneralForceElements
// Children: None
// Purpose:  Applies torsional spring and damper between two pinned bodies.
// Author:   Sanjay Dastoor - 06/13/07
//-----------------------------------------------------------------------------
#ifndef __USERFORCETORSIONALSPRING_H__
#define __USERFORCETORSIONALSPRING_H__
//-----------------------------------------------------------------------------
#include "StandardCppHeadersAndNamespace.h"
#include "SimTKsimbody.h"
using namespace SimTK;

//--------------------------------------------------------------------------
// User-defined classes for adding forces/torques are constructed as follows:
// 1. Create a constructor with whatever arguments make sense for the force or torque and copy the arguments into class data.
// 2. Create a clone method (all clone methods are identical except for the class name appearing after "new")
// 3. Create a calc  method (the arguments and return type for all calc methods are identical).
//    The code in the calc method is specific to the calculation of force or torque.
// Note: The set of all forces is replaced by an equivalent set, consisting of a torque
//       that is equal to the moment of the forces about the body's origin together
//       with the resultant of the forces applied at the body's origin.
//--------------------------------------------------------------------------
class UserForceTorsionalSpring : public GeneralForceElements::UserForce 
{
public:
	// Constructor is explicit
	explicit UserForceTorsionalSpring( BodyId bodyIdA, BodyId bodyIdB, Real springConstant, Real x0)
	{ 
		myBodyIdA = bodyIdA;
		myBodyIdB = bodyIdB;
		mySpringConstant = springConstant;
		myX0 = x0;
	} 

	// The clone method is used internally by Simbody (required by virtual parent class)
	UserForce*  clone() const  { return new UserForceTorsionalSpring(*this); }

	// The calc method is where forces or torques are calculated (required by virtual parent class)
	void  calc( const MatterSubsystem& matter,              // Input information (matter)
		const State&           state,               // Input information (current state)
		Vector_<SpatialVec>&   bodyForces,          // Forces and torques on bodies
		Vector_<Vec3>&         particleForces,      // Forces on particles (currently unused)
		Vector&                mobilityForces,      // Generalized forces 
		Real&                  pe ) const           // For forces with a potential energy
	{
		Real angleA = matter.getMobilizerQ(state,myBodyIdA,0);
		Real angleB = matter.getMobilizerQ(state,myBodyIdB,0);
		Real currentAngle = angleA - angleB;
		Real torqueH = -mySpringConstant*(currentAngle-myX0);

		// Get the proper memory location to increment the force and/or torque.
		// bodiesForces is a Vec6 whose elements are two Vec3.
		// The elements of the first  Vec3 are Tx, Ty, Tz  (expressed in the ground's "x,y,z").
		// The elements of the second Vec3 are Fx, Fy, Fz  (expressed in the ground's "x,y,z").
		SpatialVec& bodiesForces1 = bodyForces[ myBodyIdA];
		Vec3& torqueSum1 = bodiesForces1[0];
		Vec3& forceSum1  = bodiesForces1[1];

		// Increment the sum of all forces on this body (other force subsystems may also add forces/torque)
//		 torqueSum1[0] += torqueH;       // Increment torque in the ground's x-direction./
//		 torqueSum1[1] += torqueH;       // Increment torque in the ground's y-direction.
		torqueSum1[2] += torqueH;       // Increment torque in the ground's z-direction.
		// forceSum[0]  += 0;       // Increment force  in the ground's x-direction.
		// forceSum[1]  += 0;       // Increment force  in the ground's y-direction.
		// forceSum[0]  += 0;       // Increment force  in the ground's z-direction.

		// Get the proper memory location to increment the force and/or torque.
		// bodiesForces is a Vec6 whose elements are two Vec3.
		// The elements of the first  Vec3 are Tx, Ty, Tz  (expressed in the ground's "x,y,z").
		// The elements of the second Vec3 are Fx, Fy, Fz  (expressed in the ground's "x,y,z").
		SpatialVec& bodiesForces2 = bodyForces[ myBodyIdB];
		Vec3& torqueSum2 = bodiesForces2[0];
		Vec3& forceSum2  = bodiesForces2[1];


		// Increment the sum of all forces on this body (other force subsystems may also add forces/torque)
		// torqueSum[0] += 0;       // Increment torque in the ground's x-direction.
		// torqueSum[1] += 0;       // Increment torque in the ground's y-direction.
		torqueSum2[2] -= torqueH;       // Increment torque in the ground's z-direction.
		// forceSum[0]  += 0;       // Increment force  in the ground's x-direction.
		// forceSum[1]  += 0;       // Increment force  in the ground's y-direction.
		// forceSum[0]  += 0;       // Increment force  in the ground's z-direction.
	}

private:
	BodyId  myBodyIdA;
	BodyId  myBodyIdB;
	Real    mySpringConstant;
	Real    myX0;
};
//-----------------------------------------------------------------------------
#endif /* __USERFORCETORSIONALSPRING_H__ */
//-----------------------------------------------------------------------------
